/*
 * Copyright (C) 2006 by Inlab Software GmbH, Gruenwald, Germany
 *         All Rights Reserved / Alle Rechte vorbehalten
 *
 * Use of and access to this source code requires either a written 
 * permission of Inlab Software GmbH (source code license) or has 
 * to be covered by a valid written contractual agreement. 
 * 
 * $RCSfile: bngagent.c,v $
 * $Revision: 1.44 $
 * $Date: 2006/11/29 18:00:21 $
 */

#include <stdio.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/termios.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <net/if.h>
#include <stdlib.h>
#include <unistd.h>
#include <resolv.h>
#include <fcntl.h>
#include <netdb.h>
#include <string.h>
#include <strings.h>
#include <signal.h>
#include <sysexits.h>
#include <syslog.h>

#define  MAXBYTES 64 
#define  MAXLINE  64 

static int foreground=0;
static int debugflag=0;
static int testflag=0;

#ifdef __hpux
#include <sys/pstat.h>

getloadavg (double *ave,int na) {
  struct pst_dynamic pstd;
  int i;

  if (pstat_getdynamic(&pstd,sizeof(struct pst_dynamic),(size_t)1,0)<0) {
    for (i=0; i<na; i++) {
      *(ave + i) = 0.0;
    }
    if(debugflag) {
      fprintf(stderr, "pstat_getdynamic() failed, assuming 0.0\n"); 
    }
    return (-1);
  }

  switch (na) {
    case 3:
      *(ave + 2) = pstd.psd_avg_15_min;
    case 2:
      *(ave + 1) = pstd.psd_avg_5_min;
    case 1:
      *(ave) = pstd.psd_avg_1_min;
    default:
      break;
  }
  return (3);
}
#else
int getloadavg(double*, int);
#endif

void usage() {
  fprintf(stderr,"\n");
  fprintf(stderr,"  $Revision: 1.44 $\n");
  fprintf(stderr,"  bngagent is an open source part of the BalanceNG product\n");
  fprintf(stderr,"  Copyright (C) 2006 by Inlab Software GmbH, Gruenwald, Germany\n");
  fprintf(stderr,"  All rights reserved - more infos at: http://www.BalanceNG.net\n");
  fprintf(stderr,"\n");
  fprintf(stderr,"  usage:\n");
  fprintf(stderr,"    server            bngagent <options>    port\n");
  fprintf(stderr,"    request (test)    bngagent <options> -r address:port\n");
  fprintf(stderr,"  options:\n");
  fprintf(stderr,"    -0                return 1 minute load avg    (server,default)\n");
  fprintf(stderr,"    -1                return 5 minute load avg    (server)\n");
  fprintf(stderr,"    -2                return 15 minute load avg   (server)\n");
  fprintf(stderr,"    -f                stay in foreground          (server)\n");
  fprintf(stderr,"    -b <address>      specify bind address        (both)\n");
  fprintf(stderr,"    -c <command>      specify command             (server)\n");
  fprintf(stderr,"    -d                enable debug and foreground (both)\n");
  fprintf(stderr,"    -t <targetid>     specify targetid            (request)\n");
  fprintf(stderr,"\n");
  exit(EX_USAGE);
}

static void daemonize() {
  int fd,childpid;

  if ((childpid = fork()) < 0) {
    fprintf(stderr,"cannot fork\n");
    exit(EX_OSERR);
  } else {
    if (childpid > 0) {
      exit(EX_OK); 
    }
  }
  #ifdef __hpux 
  (void) setsid();
  #else
  #ifdef BSD 
    setpgid(getpid(), 0);
  #else 
    setpgrp();
  #endif
  {
    int fd;
    if((fd = open("/dev/tty", O_RDWR)) != -1) {
      if(ioctl(fd, TIOCNOTTY) < 0) {
        fprintf(stderr, "error: cannot ioctl(IOCNOTTY)");
        exit(EX_OSERR);
      }
      close(fd);
    }
  }
  #endif

  openlog("bngagent", LOG_ODELAY | LOG_PID | LOG_CONS, LOG_DAEMON);

  if((fd=open("/dev/null", O_RDWR, 0)) < 0) {
    syslog(LOG_CRIT, "unable to open /dev/null");
    exit(EX_OSERR);
  }

  if(dup2(fd, 0) < 0)
    syslog(LOG_ERR, "Unable to set /dev/null as stdin");
  if(dup2(fd, 1) < 0)
    syslog(LOG_ERR, "Unable to set /dev/null as stdout");
  if(dup2(fd, 2) < 0)
    syslog(LOG_ERR, "Unable to set /dev/null as stderr");
  close(fd);

  chdir("/");
}

static char*  	bindhost=NULL;
static char*  	command=NULL;
static char*  	targetid=NULL;
static struct 	sockaddr_in addr, tr_sockaddr;
static int	sd, tr_sockaddr_len;
static struct   in_addr tr_ipaddr;
static int      tr_port;
static int      laindex=0;	

static void prepare_testrequest(char* destination) {
  struct hostent* he;
  struct servent* sp;
  char   *server_cp, *port_cp;

  server_cp = strtok(destination, ":");
  port_cp = strtok(NULL, ":");
  if(server_cp == NULL || port_cp == NULL) {
    usage();
  }
  he=gethostbyname(server_cp);
  if(he == NULL) {
    if ((tr_ipaddr.s_addr = inet_addr(server_cp)) == (unsigned int) -1) {
      fprintf(stderr, "unknown or invalid IP address [%s]\n", server_cp);
      usage();
    }
  } else {
    memcpy(&tr_ipaddr, he->h_addr, he->h_length);
  }
  sp = getservbyname(port_cp,"udp");
  if(sp == NULL) {
    tr_port=atoi(port_cp);
  } else {
    tr_port=ntohs(sp->s_port);
  }
  if(tr_port == 0) {
    fprintf(stderr, "unknown or invalid port [%s]\n", port_cp);
    usage();
  }
}

static void testrequest_timeout() {
  printf("*timeout*\n");
  exit(EX_OK);
}

static void perform_testrequest() {
  int rc;
  static unsigned char buffer[MAXBYTES];
  unsigned char intbuffer[4];

  tr_sockaddr.sin_port = htons(tr_port);
  tr_sockaddr.sin_family = AF_INET;
  memcpy(&tr_sockaddr.sin_addr,&tr_ipaddr,sizeof(tr_sockaddr.sin_addr));
  bzero(buffer, MAXBYTES);
  if(targetid != NULL) {
    *((unsigned short*) buffer) = htons(atoi(targetid));
  }
  rc=sendto(sd, buffer, sizeof(unsigned short), 0, 
         (struct sockaddr*)&tr_sockaddr, sizeof(tr_sockaddr));
  if(rc < 0) { 
    perror("sendto");
    exit(EX_OSERR);
  }

  signal(SIGALRM,testrequest_timeout);
  alarm(1);
  tr_sockaddr_len=sizeof(tr_sockaddr);
  rc = recvfrom(sd, buffer, sizeof(buffer), 0, 
         (struct sockaddr*)&tr_sockaddr, &tr_sockaddr_len);

  if(rc == 6) {
    memcpy(intbuffer, buffer+sizeof(unsigned short), sizeof(unsigned int));
    printf("target id: %u value: %u\n", ntohs(*((unsigned short*) buffer)), 
           ntohl(*((unsigned int*) intbuffer)));  
  } else {
    printf("invalid length %d returned\n", rc);
  }
  exit(EX_OK);
}

int main(int argc, char** argv) {
  int 	 c,port; 
  struct servent* sp;

  unsigned char buffer[MAXBYTES];
  unsigned char intbuffer[4];

  while((c = getopt(argc,argv,"012b:fdc:rt:")) != EOF) {
    switch(c) {
    case 'b':
      bindhost=strdup(optarg);
      break;
    case 'f':
      foreground=1;
      break;
    case 'd':
      debugflag=1;
      foreground=1;
      break;
    case 'c':
      command=strdup(optarg);
      break;
    case 't':
      targetid=strdup(optarg);
      break;
    case 'r':
      testflag=1;
      break;
    case '0':
      laindex=0;
      break;
    case '1':
      laindex=1;
      break;
    case '2':
      laindex=2;
      break;
    case '?':
    default:
      usage();
    }
  }
  argc -= optind;
  argv += optind;

  if(argc != 1) {
    usage();
  }     

  if(testflag) {
    prepare_testrequest(argv[0]); 	/* prepare test request */
  }
  
  if(testflag) {
    port=0; 				/* port will be reassigned */
  } else {
    sp=getservbyname(argv[0],"udp");   
    if(sp==NULL) {
      port=atoi(argv[0]); 
      if(port==0) {
        fprintf(stderr, "error: invalid port specifier\n");
        exit(EX_USAGE);
      }
    } else {
      port=ntohs(sp->s_port);
    }
  }

  if(debugflag) 
    fprintf(stderr,"we will listen on port %d\n",port);

  sd=socket(PF_INET, SOCK_DGRAM, 0);
  bzero(&addr, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);

  if(bindhost == NULL) {
    addr.sin_addr.s_addr = INADDR_ANY;
  } else {
    struct hostent *he;
    he = gethostbyname(bindhost);
    if(he == NULL) {
      int taddr;
      if((taddr = inet_addr(bindhost)) == -1) {
        fprintf(stderr,"unknown or invalid address [%s]\n",bindhost);
        exit(EX_USAGE);
      } else {
        addr.sin_addr.s_addr = taddr;
      }
    } else {
      memcpy(&addr.sin_addr, he->h_addr, he->h_length);
    }
  }

  if(bind(sd, (struct sockaddr*)&addr, sizeof(addr)) != 0) {
    perror("bind");
    exit(EX_OSERR);
  }
 
  if(testflag) {
    perform_testrequest();
    exit(EX_OK);
  }

  if(!foreground)
    daemonize();

  for(;;) {	
    int  bytes, addr_len=sizeof(addr);
    unsigned int returnvalue;
    FILE *pipe;
    char line[MAXLINE];

    bytes = recvfrom(sd, buffer, sizeof(buffer), 0, 
                     (struct sockaddr*)&addr, &addr_len);

    if(debugflag) { 
      fprintf(stderr,"packet from %s:%d (%d bytes) target %d\n", 
        inet_ntoa(addr.sin_addr), 
	ntohs(addr.sin_port), 
	bytes, 
	ntohs(*((unsigned short*) buffer)));
    }

    if(command != NULL) {
      if(debugflag) 
        fprintf(stderr,"executing script or command [%s]\n", command);
      if((pipe=popen(command,"r")) == NULL) {
        fprintf(stderr,"error popen()\n");
        exit(EX_OSERR);
      }
      if(fgets(line,MAXLINE,pipe) == NULL) {
        fprintf(stderr,"error fgets()\n"); 
        exit(EX_OSERR);
      }
      pclose(pipe);
      returnvalue=atoi(line);
    } else {
      double values[3];
      if(getloadavg(values, 3) != 3) {
	values[0]=0.0;
	values[1]=0.0;
	values[2]=0.0;
      }
      if(debugflag) {
	if(laindex==0)
          fprintf(stderr,"1-minute system-load: %f\n",values[0]);
	if(laindex==1)
          fprintf(stderr,"5-minute system-load: %f\n",values[1]);
	if(laindex==2)
          fprintf(stderr,"15-minute system-load: %f\n",values[2]);
      }
      if(values[laindex] < 0.0) 
	values[laindex]=0.0;
      returnvalue=(unsigned int) (values[laindex]*100.0);

      /* 0 would mean service unavailable, so we return at least 1 */
      if(returnvalue == 0) 
        returnvalue++; 
    }

    if(debugflag) 
      fprintf(stderr,"about to return %d\n", returnvalue);

    /* avoiding alignment problems */ 
    returnvalue = htonl(returnvalue);
    *((unsigned int*) intbuffer)=returnvalue;	

    memcpy(buffer+sizeof(unsigned short), intbuffer, sizeof(unsigned int));

    sendto(sd, buffer, sizeof(unsigned short)+sizeof(unsigned int), 0, 
           (struct sockaddr*)&addr, sizeof(addr));
  }
}

